.tabs{display:flex;flex-direction:row;border-bottom:1px solid #e9e9e9}.tabs .tab{padding:10px 20px 10px 0;cursor:pointer;font-weight:300}.selected:is(.tabs .tab){border-bottom:solid 1px black;font-weight:400}main[data-v-b09113ad]{padding:0 5vw;font-size:12px}.dataframe[data-v-b09113ad]{margin-top:10px}.images[data-v-b09113ad]{display:grid;grid-template-columns:1fr 1fr 1fr 1fr;place-items:center center}.buttons>p[data-v-b09113ad]{padding:10px}.toasts[data-v-b09113ad]{padding-top:10px}.toasts[data-v-b09113ad]>*{margin-bottom:10px}.text-inputs>p[data-v-b09113ad]{padding:10px}.editable-list-tab[data-v-b09113ad]{display:grid;padding-top:10px;grid-template-columns:1fr 1fr}.editable-list-container[data-v-b09113ad]{display:block;width:33vw;height:300px}.editable-list-container .editable-list-container-scrollable[data-v-b09113ad]{height:100%;overflow-y:auto}.icons[data-v-b09113ad]{display:grid;padding-top:10px;gap:20px;grid-template-columns:1fr 1fr 1fr 1fr}.icons>div[data-v-b09113ad]{display:flex;gap:10px}.icons [class^=icon-][data-v-b09113ad],.icons [class*=" icon-"][data-v-b09113ad]{color:var(--color-icon-primary);font-size:20px}.draggable-list-container[data-v-b09113ad]{max-height:80vh;margin-top:10px}.drag-me[data-v-b09113ad]{width:200px;padding:10px;margin-top:10px;background-color:var(--color-background-secondary);cursor:move;-webkit-user-select:none;-moz-user-select:none;user-select:none}.floating-tooltip-tab[data-v-b09113ad]{display:grid;padding:40px;gap:40px;grid-template-columns:1fr 1fr 1fr}
